"""Client for the MADSci Experiment Manager."""

from typing import Optional, Union

import requests
from madsci.common.context import get_current_madsci_context
from madsci.common.types.experiment_types import (
    Experiment,
    ExperimentalCampaign,
    ExperimentDesign,
    ExperimentRegistration,
    ExperimentStatus,
)
from pydantic import AnyUrl
from ulid import ULID


class ExperimentClient:
    """Client for the MADSci Experiment Manager."""

    experiment_server_url: AnyUrl

    def __init__(
        self,
        experiment_server_url: Optional[Union[str, AnyUrl]] = None,
    ) -> "ExperimentClient":
        """
        Create a new Experiment Client.

        Args:
            experiment_server_url: The URL of the experiment server. If not provided, will use the URL from the current MADSci context.
        """
        self.experiment_server_url = (
            AnyUrl(experiment_server_url)
            if experiment_server_url
            else get_current_madsci_context().experiment_server_url
        )
        if not self.experiment_server_url:
            raise ValueError(
                "No experiment server URL provided, please specify a URL or set the context."
            )

    def get_experiment(self, experiment_id: Union[str, ULID]) -> dict:
        """Get an experiment by ID."""
        response = requests.get(
            f"{self.experiment_server_url}experiment/{experiment_id}", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def get_experiments(self, number: int = 10) -> list[Experiment]:
        """Get a list of the latest experiments."""
        response = requests.get(
            f"{self.experiment_server_url}experiments",
            params={"number": number},
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return [Experiment.model_validate(experiment) for experiment in response.json()]

    def start_experiment(
        self,
        experiment_design: ExperimentDesign,
        run_name: Optional[str] = None,
        run_description: Optional[str] = None,
    ) -> Experiment:
        """Start an experiment based on an ExperimentDesign."""
        response = requests.post(
            f"{self.experiment_server_url}experiment",
            json=ExperimentRegistration(
                experiment_design=experiment_design.model_dump(mode="json"),
                run_name=run_name,
                run_description=run_description,
            ).model_dump(mode="json"),
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def end_experiment(
        self, experiment_id: Union[str, ULID], status: Optional[ExperimentStatus] = None
    ) -> Experiment:
        """End an experiment by ID. Optionally, set the status."""
        response = requests.post(
            f"{self.experiment_server_url}experiment/{experiment_id}/end",
            params={"status": status},
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def continue_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Continue an experiment by ID."""
        response = requests.post(
            f"{self.experiment_server_url}experiment/{experiment_id}/continue",
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def pause_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Pause an experiment by ID."""
        response = requests.post(
            f"{self.experiment_server_url}experiment/{experiment_id}/pause", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def cancel_experiment(self, experiment_id: Union[str, ULID]) -> Experiment:
        """Cancel an experiment by ID."""
        response = requests.post(
            f"{self.experiment_server_url}experiment/{experiment_id}/cancel",
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return Experiment.model_validate(response.json())

    def register_campaign(self, campaign: ExperimentalCampaign) -> ExperimentalCampaign:
        """Register a new experimental campaign."""
        response = requests.post(
            f"{self.experiment_server_url}campaign",
            json=campaign.model_dump(mode="json"),
            timeout=10,
        )
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def get_campaign(self, campaign_id: str) -> ExperimentalCampaign:
        """Get an experimental campaign by ID."""
        response = requests.get(
            f"{self.experiment_server_url}campaign/{campaign_id}", timeout=10
        )
        if not response.ok:
            response.raise_for_status()
        return response.json()



"""MADSci Event Handling."""

import contextlib
import inspect
import json
import logging
import queue
import time
import traceback
import warnings
from collections import OrderedDict
from pathlib import Path
from threading import Lock, Thread
from typing import Any, Optional, Union

import requests
from madsci.common.context import get_current_madsci_context
from madsci.common.types.event_types import (
    Event,
    EventClientConfig,
    EventType,
)
from madsci.common.utils import threaded_task
from pydantic import BaseModel, ValidationError
from rich.logging import RichHandler


class EventClient:
    """A logger and event handler for MADSci system components."""

    config: Optional[EventClientConfig] = None
    _event_buffer = queue.Queue()
    _buffer_lock = Lock()
    _retry_thread = None
    _retrying = False

    def __init__(
        self,
        config: Optional[EventClientConfig] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the event logger. If no config is provided, use the default config.

        Keyword Arguments are used to override the values of the passed in/default config.
        """
        if kwargs:
            self.config = (
                EventClientConfig(**kwargs)
                if not config
                else config.model_copy(update=kwargs)
            )
        else:
            self.config = config or EventClientConfig()
        if self.config.name:
            self.name = self.config.name
        else:
            # * See if there's a calling module we can name after
            stack = inspect.stack()
            parent = stack[1][0]
            if calling_module := parent.f_globals.get("__name__"):
                self.name = calling_module
            else:
                # * No luck, name after EventClient
                self.name = __name__
        self.name = str(self.name)
        self.logger = logging.getLogger(self.name)
        self.log_dir = Path(self.config.log_dir).expanduser()
        self.log_dir.mkdir(parents=True, exist_ok=True)
        self.logfile = self.log_dir / f"{self.name}.log"
        self.logger.setLevel(self.config.log_level)
        for handler in self.logger.handlers:
            self.logger.removeHandler(handler)
        file_handler = logging.FileHandler(filename=str(self.logfile), mode="a+")
        self.logger.addHandler(file_handler)
        self.logger.addHandler(RichHandler(rich_tracebacks=True, show_path=False))
        self.event_server = (
            self.config.event_server_url
            or get_current_madsci_context().event_server_url
        )

    def get_log(self) -> dict[str, Event]:
        """Read the log"""
        events = {}
        with self.logfile.open() as log:
            for line in log.readlines():
                try:
                    event = Event.model_validate_json(line)
                except ValidationError:
                    event = Event(event_type=EventType.UNKNOWN, event_data=line)
                events[event.event_id] = event
        return events

    def get_event(self, event_id: str) -> Optional[Event]:
        """Get a specific event by ID."""
        if self.event_server:
            response = requests.get(
                str(self.event_server) + f"event/{event_id}",
                timeout=10,
            )
            if not response.ok:
                response.raise_for_status()
            return Event.model_validate(response.json())
        events = self.get_log()
        return events.get(event_id, None)

    def get_events(self, number: int = 100, level: int = -1) -> dict[str, Event]:
        """Query the event server for a certain number of recent events. If no event server is configured, query the log file instead."""
        if level == -1:
            level = int(self.logger.getEffectiveLevel())
        events = OrderedDict()
        if self.event_server:
            response = requests.get(
                str(self.event_server) + "events",
                timeout=10,
                params={"number": number, "level": level},
            )
            if not response.ok:
                response.raise_for_status()
            for key, value in response.json().items():
                events[key] = Event.model_validate(value)
            return dict(events)
        events = self.get_log()
        selected_events = {}
        for event in reversed(list(events.values())):
            selected_events[event.event_id] = event
            if len(selected_events) >= number:
                break
        return selected_events

    def query_events(self, selector: dict) -> dict[str, Event]:
        """Query the event server for events based on a selector. Requires an event server be configured."""
        events = OrderedDict()
        if self.event_server:
            response = requests.post(
                str(self.event_server) + "events/query",
                timeout=10,
                params={"selector": selector},
            )
            if not response.ok:
                response.raise_for_status()
            for key, value in response.json().items():
                events[key] = Event.model_validate(value)
            return dict(events)
        self.logger.warning("No event server configured. Cannot query events.")
        return {}

    def log(
        self,
        event: Union[Event, Any],
        level: Optional[int] = None,
        alert: Optional[bool] = None,
        warning_category: Optional[Warning] = None,
    ) -> None:
        """Log an event."""
        # * If we've got a string or dict, check if it's a serialized event
        if isinstance(event, str):
            with contextlib.suppress(ValidationError):
                event = Event.model_validate_json(event)
        if isinstance(event, dict):
            with contextlib.suppress(ValidationError):
                event = Event.model_validate(event)
        if isinstance(event, Exception):
            event = Event(
                event_type=EventType.LOG_ERROR,
                event_data=traceback.format_exc(),
                log_level=logging.ERROR,
            )
        if not isinstance(event, Event):
            event = self._new_event_for_log(event, level)
        event.log_level = level if level is not None else event.log_level
        event.alert = alert if alert is not None else event.alert
        if warning_category and self.logger.getEffectiveLevel() <= logging.WARNING:
            # * Warn via the warnings module
            warnings.warn(
                event.event_data,
                category=warning_category,
                stacklevel=3,
            )
        else:
            self.logger.log(level=event.log_level, msg=event.model_dump_json())
        # * Log the event to the event server if configured
        # * Only log if the event is at the same level or higher than the logger
        if self.logger.getEffectiveLevel() <= event.log_level and self.event_server:
            self._send_event_to_event_server_task(event)

    def log_debug(self, event: Union[Event, str]) -> None:
        """Log an event at the debug level."""
        self.log(event, logging.DEBUG)

    debug = log_debug

    def log_info(self, event: Union[Event, str]) -> None:
        """Log an event at the info level."""
        self.log(event, logging.INFO)

    info = log_info

    def log_warning(
        self, event: Union[Event, str], warning_category: Warning = UserWarning
    ) -> None:
        """Log an event at the warning level."""
        self.log(event, logging.WARNING, warning_category=warning_category)

    warning = log_warning
    warn = log_warning

    def log_error(self, event: Union[Event, str]) -> None:
        """Log an event at the error level."""
        self.log(event=event, level=logging.ERROR)

    error = log_error

    def log_critical(self, event: Union[Event, str]) -> None:
        """Log an event at the critical level."""
        self.log(event, logging.CRITICAL)

    critical = log_critical

    def log_alert(self, event: Union[Event, str]) -> None:
        """Log an event at the alert level."""
        self.log(event, alert=True)

    alert = log_alert

    def get_utilization_periods(
        self,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        analysis_type: str = "daily",
        user_timezone: str = "America/Chicago",
        include_users: bool = True,
        csv_export: bool = False,
        save_to_file: bool = False,
        output_path: Optional[str] = None,
    ) -> Optional[Union[dict[str, Any], str]]:
        """
        Get time-series utilization analysis with periodic breakdowns, optionally export to CSV.

        Args:
            start_time: ISO format start time
            end_time: ISO format end time
            analysis_type: "hourly", "daily", "weekly", "monthly"
            user_timezone: Timezone for day boundaries (e.g., "America/Chicago")
            include_users: Whether to include user utilization data
            csv_export: If True, convert report to CSV format
            save_to_file: If True, save to file (requires output_path)
            output_path: Path to save files (used when save_to_file=True)

        Returns:
            - If csv_export=False: JSON dict with utilization data
            - If csv_export=True and save_to_file=False: CSV string
            - If csv_export=True and save_to_file=True: dict with file save results
        """
        if not self.event_server:
            self.logger.warning("No event server configured.")
            return None

        try:
            params = {
                "analysis_type": analysis_type,
                "user_timezone": user_timezone,
                "include_users": str(include_users).lower(),
            }
            if start_time:
                params["start_time"] = start_time
            if end_time:
                params["end_time"] = end_time

            if csv_export:
                params["csv_format"] = "true"
                if save_to_file and output_path:
                    params["save_to_file"] = "true"
                    params["output_path"] = output_path

            response = requests.get(
                str(self.event_server) + "utilization/periods",
                params=params,
                timeout=30,
            )
            if not response.ok:
                self.logger.error(
                    f"Error getting utilization periods: HTTP {response.status_code}"
                )
                response.raise_for_status()

            # Handle CSV response - check if content type contains 'text/csv'
            content_type = response.headers.get("content-type", "").lower()
            if csv_export and "text/csv" in content_type:
                return response.text

            # Handle JSON response (either regular JSON or file save results)
            return response.json()

        except requests.RequestException as e:
            self.logger.error(f"Error getting utilization periods: {e}")
            return None

    def get_session_utilization(
        self,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        csv_export: bool = False,
        save_to_file: bool = False,
        output_path: Optional[str] = None,
    ) -> Optional[Union[dict[str, Any], str]]:
        """
        Get session-based utilization report, optionally export to CSV.

        Sessions represent workcell/lab start and stop periods. Each session
        indicates when laboratory equipment was actively configured and available.

        Args:
            start_time: ISO format start time
            end_time: ISO format end time
            csv_export: If True, convert report to CSV format
            save_to_file: If True, save to file (requires output_path)
            output_path: Path to save files (used when save_to_file=True)

        Returns:
            - If csv_export=False: JSON dict
            - If csv_export=True and save_to_file=False: CSV string
            - If csv_export=True and save_to_file=True: dict with file save results
        """
        if not self.event_server:
            return None

        try:
            params = {}
            if start_time:
                params["start_time"] = start_time
            if end_time:
                params["end_time"] = end_time

            if csv_export:
                params["csv_format"] = "true"
                if save_to_file and output_path:
                    params["save_to_file"] = "true"
                    params["output_path"] = output_path

            response = requests.get(
                str(self.event_server) + "utilization/sessions",
                params=params,
                timeout=60,
            )

            if not response.ok:
                response.raise_for_status()

            # Handle CSV response - check if content type contains 'text/csv'
            content_type = response.headers.get("content-type", "").lower()
            if csv_export and "text/csv" in content_type:
                return response.text

            # Handle JSON response (either regular JSON or file save results)
            return response.json()

        except requests.RequestException:
            return None

    def get_user_utilization_report(
        self,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        csv_export: bool = False,
        save_to_file: bool = False,
        output_path: Optional[str] = None,
    ) -> Optional[Union[dict[str, Any], str]]:
        """
        Get detailed user utilization report from the event server, optionally export to CSV.

        Args:
            start_time: ISO format start time (e.g., "2025-07-20T00:00:00Z")
            end_time: ISO format end time (e.g., "2025-07-23T00:00:00Z")
            csv_export: If True, convert report to CSV format
            save_to_file: If True, save to file (requires output_path)
            output_path: Path to save files (used when save_to_file=True)

        Returns:
            - If csv_export=False: JSON dict with detailed user utilization data
            - If csv_export=True and save_to_file=False: CSV string
            - If csv_export=True and save_to_file=True: dict with file save results
        """
        if not self.event_server:
            self.logger.warning("No event server configured.")
            return None

        try:
            params = {}
            if start_time:
                params["start_time"] = start_time
            if end_time:
                params["end_time"] = end_time

            if csv_export:
                params["csv_format"] = "true"
                if save_to_file and output_path:
                    params["save_to_file"] = "true"
                    params["output_path"] = output_path

            response = requests.get(
                str(self.event_server) + "utilization/users",
                params=params,
                timeout=60,
            )

            if not response.ok:
                self.logger.error(
                    f"Error getting user utilization report: HTTP {response.status_code}"
                )
                response.raise_for_status()

            # Handle CSV response - check if content type contains 'text/csv'
            content_type = response.headers.get("content-type", "").lower()
            if csv_export and "text/csv" in content_type:
                return response.text

            # Handle JSON response (either regular JSON or file save results)
            return response.json()

        except requests.RequestException as e:
            self.logger.error(f"Error getting user utilization report: {e}")
            return None

    def _start_retry_thread(self) -> None:
        with self._buffer_lock:
            if not self._retrying:
                self._retrying = True
                self._retry_thread = Thread(
                    target=self._retry_buffered_events, daemon=True
                )
                self._retry_thread.start()

    def _retry_buffered_events(self) -> None:
        backoff = 2
        max_backoff = 60
        while not self._event_buffer.empty():
            try:
                event = self._event_buffer.get()
                self._send_event_to_event_server(event, retrying=True)
                backoff = 2  # Reset backoff on success
            except Exception:
                time.sleep(backoff)
                backoff = min(backoff * 2, max_backoff)
                self._event_buffer.put(event)  # Re-add the event to the buffer
        with self._buffer_lock:
            self._retrying = False

    @threaded_task
    def _send_event_to_event_server_task(
        self, event: Event, retrying: bool = False
    ) -> None:
        """Send an event to the event manager. Buffer on failure."""
        try:
            self._send_event_to_event_server(event, retrying=retrying)
        except Exception as e:
            self.logger.error(f"Error in _send_event_to_event_server_task: {e}")

    def _send_event_to_event_server(self, event: Event, retrying: bool = False) -> None:
        """Send an event to the event manager. Buffer on failure."""

        try:
            response = requests.post(
                url=f"{self.event_server}event",
                json=event.model_dump(mode="json"),
                timeout=10,
            )

            if not response.ok:
                response.raise_for_status()

        except Exception:
            if not retrying:
                self._event_buffer.put(event)
                self._start_retry_thread()
            else:
                # If already retrying, just re-raise to trigger backoff
                raise

    def _new_event_for_log(self, event_data: Any, level: int) -> Event:
        """Create a new log event from arbitrary data"""
        event_type = EventType.LOG
        if level == logging.DEBUG:
            event_type = EventType.LOG_DEBUG
        elif level == logging.INFO:
            event_type = EventType.LOG_INFO
        elif level == logging.WARNING:
            event_type = EventType.LOG_WARNING
        elif level == logging.ERROR:
            event_type = EventType.LOG_ERROR
        elif level == logging.CRITICAL:
            event_type = EventType.LOG_CRITICAL
        if isinstance(event_data, BaseModel):
            event_data = event_data.model_dump(mode="json")
        elif isinstance(event_data, dict):
            # Keep dict as-is
            pass
        else:
            try:
                event_data = json.dumps(event_data, default=str)
            except Exception:
                try:
                    event_data = str(event_data)
                except Exception:
                    event_data = {
                        "error": "Error during logging. Unable to serialize event data."
                    }
        return Event(
            event_type=event_type,
            event_data=event_data,
        )



"""REST-based node client implementation."""

import tempfile
import time
import zipfile
from pathlib import Path
from typing import Any, ClassVar, Optional, Union

import requests
from madsci.client.event_client import EventClient
from madsci.client.node.abstract_node_client import (
    AbstractNodeClient,
)
from madsci.common.types.action_types import (
    ActionFiles,
    ActionRequest,
    ActionResult,
    ActionStatus,
)
from madsci.common.types.admin_command_types import AdminCommandResponse
from madsci.common.types.event_types import Event
from madsci.common.types.node_types import (
    AdminCommands,
    NodeClientCapabilities,
    NodeInfo,
    NodeSetConfigResponse,
    NodeStatus,
)
from madsci.common.types.resource_types import ResourceDataModels
from pydantic import AnyUrl


def _serialize_for_json(obj: Any) -> Any:
    """
    Recursively serialize Pydantic models and other complex objects for JSON transmission.

    Args:
        obj: The object to serialize

    Returns:
        A JSON-serializable representation of the object
    """
    if hasattr(obj, "model_dump"):
        # This is a Pydantic model - use model_dump(mode="json") for proper serialization
        return obj.model_dump(mode="json")
    if isinstance(obj, dict):
        # Recursively serialize dictionary values
        return {key: _serialize_for_json(value) for key, value in obj.items()}
    if isinstance(obj, (list, tuple)):
        # Recursively serialize list/tuple items
        return [_serialize_for_json(item) for item in obj]
    # For primitive types and other objects, return as-is
    return obj


class RestNodeClient(AbstractNodeClient):
    """REST-based node client."""

    url_protocols: ClassVar[list[str]] = ["http", "https"]
    """The protocols supported by this client."""

    supported_capabilities: NodeClientCapabilities = NodeClientCapabilities(
        # *Supported capabilities
        get_info=True,
        get_state=True,
        get_status=True,
        send_action=True,
        get_action_status=True,
        get_action_result=True,
        get_action_history=True,
        action_files=True,
        send_admin_commands=True,
        set_config=True,
        get_log=True,
        # *Unsupported Capabilities
        get_resources=False,
    )

    def __init__(self, url: AnyUrl) -> "RestNodeClient":
        """Initialize the client."""
        super().__init__(url)
        self.logger = EventClient()

    def send_action(
        self,
        action_request: ActionRequest,
        await_result: bool = True,
        timeout: Optional[float] = None,
    ) -> ActionResult:
        """Perform the action defined by action_request on the specified node."""
        try:
            # Step 1: Create the action
            action_id = self._create_action(action_request)

            # Step 2: Upload files if any
            if action_request.files:
                self._upload_action_files(
                    action_request.action_name, action_id, action_request.files
                )

            # Step 3: Start the action
            result = self._start_action(action_request.action_name, action_id)

            # Step 4: Wait for completion if requested
            if await_result and not result.status.is_terminal:
                result = self.await_action_result_by_name(
                    action_request.action_name, action_id, timeout=timeout
                )

            return result

        except requests.HTTPError as e:
            if hasattr(e, "response") and e.response is not None:
                self.logger.error(f"{e.response.status_code}: {e.response.text}")
            else:
                self.logger.error(str(e))
            raise e

    def _create_action(self, action_request: ActionRequest) -> str:
        """Create a new action and return the action_id. REST-implementation specific"""

        # Convert ActionRequest to RestActionRequest format
        args = dict(action_request.args) if action_request.args else {}

        # Serialize all arguments to ensure Pydantic models are JSON-serializable
        serialized_args = _serialize_for_json(args)
        serialized_var_args = (
            _serialize_for_json(action_request.var_args)
            if action_request.var_args is not None
            else None
        )
        serialized_var_kwargs = (
            _serialize_for_json(action_request.var_kwargs)
            if action_request.var_kwargs is not None
            else None
        )

        request_data = {"args": serialized_args}
        if serialized_var_args is not None:
            request_data["var_args"] = serialized_var_args
        if serialized_var_kwargs is not None:
            request_data["var_kwargs"] = serialized_var_kwargs

        rest_response = requests.post(
            f"{self.url}/action/{action_request.action_name}",
            json=request_data,
            timeout=60,
        )
        rest_response.raise_for_status()
        response_data = rest_response.json()
        return response_data["action_id"]

    def _upload_action_files(
        self, action_name: str, action_id: str, files: dict[str, Union[str, list[str]]]
    ) -> None:
        """Upload files for an action. REST-implementation specific"""
        for file_key, file_value in files.items():
            if isinstance(file_value, list):
                # Handle list[Path] parameters - upload multiple files
                files_to_upload = []
                for file_path in file_value:
                    path = Path(file_path).expanduser()
                    files_to_upload.append(
                        (
                            file_key,
                            (path.name, path.open("rb"), "application/octet-stream"),
                        )
                    )

                try:
                    rest_response = requests.post(
                        f"{self.url}/action/{action_name}/{action_id}/upload/{file_key}",
                        files=files_to_upload,
                        timeout=60,
                    )
                    rest_response.raise_for_status()
                finally:
                    # Close any opened file handles
                    for _, (_, file_handle, _) in files_to_upload:
                        if hasattr(file_handle, "close"):
                            file_handle.close()
            else:
                # Handle single Path parameters
                with Path(file_value).expanduser().open("rb") as file_handle:
                    rest_response = requests.post(
                        f"{self.url}/action/{action_name}/{action_id}/upload/{file_key}",
                        files={"file": file_handle},
                        timeout=60,
                    )
                    rest_response.raise_for_status()

    def _start_action(self, action_name: str, action_id: str) -> ActionResult:
        """Start an action that has been created. REST-implementation specific."""
        rest_response = requests.post(
            f"{self.url}/action/{action_name}/{action_id}/start",
            timeout=60,
        )
        rest_response.raise_for_status()
        return ActionResult.model_validate(rest_response.json())

    def get_action_status_by_name(
        self, action_name: str, action_id: str
    ) -> ActionStatus:
        """Get the status of an action by action name."""
        rest_response = requests.get(
            f"{self.url}/action/{action_name}/{action_id}/status",
            timeout=10,
        )
        rest_response.raise_for_status()
        return ActionStatus(rest_response.json())

    def _get_result_files(
        self, action_name: str, action_id: str, result: ActionResult
    ) -> None:
        """Return the files associated with an action result. REST-implementation specific."""
        if result.files is None:
            return

        # Always download as ZIP and extract files
        zip_path = self._get_action_files_zip(action_name, action_id)

        if isinstance(result.files, ActionFiles):
            # Multiple files case - extract from ZIP
            files_dict = result.files.model_dump()
            downloaded_files = {}

            # Extract files from ZIP
            with zipfile.ZipFile(zip_path, "r") as zip_file:
                temp_dir = Path(tempfile.mkdtemp())
                zip_file.extractall(temp_dir)

                # Map extracted files back to their labels
                for file_key in files_dict:
                    # Look for file with matching label (with any extension)
                    extracted_files = list(temp_dir.glob(f"{file_key}*"))
                    if extracted_files:
                        downloaded_files[file_key] = extracted_files[0]
                    else:
                        # Fallback: use the first file if label matching fails
                        all_files = list(temp_dir.iterdir())
                        if all_files:
                            downloaded_files[file_key] = all_files[0]

            result.files = ActionFiles.model_validate(downloaded_files)

        elif isinstance(result.files, Path):
            # Single file case - extract from ZIP
            with zipfile.ZipFile(zip_path, "r") as zip_file:
                temp_dir = Path(tempfile.mkdtemp())
                zip_file.extractall(temp_dir)

                # Get the first (and likely only) file from the ZIP
                extracted_files = list(temp_dir.iterdir())
                if extracted_files:
                    result.files = extracted_files[0]

        # Clean up the ZIP file
        zip_path.unlink(missing_ok=True)

    def get_action_result_by_name(
        self, action_name: str, action_id: str, include_files: bool = True
    ) -> ActionResult:
        """Get the result of an action by name. REST-implementation specific."""
        rest_response = requests.get(
            f"{self.url}/action/{action_name}/{action_id}/result",
            timeout=10,
        )
        rest_response.raise_for_status()
        result = ActionResult.model_validate(rest_response.json())

        if result.files and include_files:
            self._get_result_files(action_name, action_id, result)

        return result

    def _get_action_files_zip(self, action_name: str, action_id: str) -> Path:
        """Download all files from an action result as a ZIP. REST-implementation specific."""
        rest_response = requests.get(
            f"{self.url}/action/{action_name}/{action_id}/download",
            timeout=60,
        )
        rest_response.raise_for_status()

        # Save to temporary file
        with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as temp_file:
            temp_file.write(rest_response.content)
            return Path(temp_file.name)

    def get_action_history(
        self, action_id: Optional[str] = None
    ) -> dict[str, list[ActionResult]]:
        """Get the history of a single action performed on the node, or every action, if no action_id is specified."""
        response = requests.get(
            f"{self.url}/action", params={"action_id": action_id}, timeout=10
        )
        response.raise_for_status()
        return response.json()

    def get_action_status(self, action_id: str) -> ActionStatus:
        """Get the status of an action on the node."""
        rest_response = requests.get(
            f"{self.url}/action/{action_id}/status",
            timeout=10,
        )
        rest_response.raise_for_status()
        return ActionStatus(rest_response.json())

    def get_action_result(self, action_id: str) -> ActionResult:
        """Get the result of an action on the node."""
        rest_response = requests.get(
            f"{self.url}/action/{action_id}/result",
            timeout=10,
        )
        rest_response.raise_for_status()
        return ActionResult.model_validate(rest_response.json())

    def await_action_result(
        self, action_id: str, timeout: Optional[float] = None
    ) -> ActionResult:
        """Wait for an action to complete and return the result. Optionally, specify a timeout in seconds."""
        start_time = time.time()
        interval = 0.25
        while True:
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out waiting for action to complete.")
            status = self.get_action_status(action_id)
            if not status.is_terminal:
                time.sleep(interval)
                interval = (
                    interval * 1.5 if interval < 5 else 5
                )  # * Capped Exponential backoff
                continue
            return self.get_action_result(action_id)

    def await_action_result_by_name(
        self, action_name: str, action_id: str, timeout: Optional[float] = None
    ) -> ActionResult:
        """Wait for an action to complete and return the result. Optionally, specify a timeout in seconds. REST-implementation specific."""
        start_time = time.time()
        interval = 0.25
        while True:
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out waiting for action to complete.")
            status = self.get_action_status_by_name(action_name, action_id)
            if not status.is_terminal:
                time.sleep(interval)
                interval = (
                    interval * 1.5 if interval < 5 else 5
                )  # * Capped Exponential backoff
                continue

            # Get the full result with data when the action is complete
            if status == ActionStatus.SUCCEEDED:
                return self.get_action_result_by_name(action_name, action_id)

            # For non-successful terminal states, create a basic ActionResult
            return ActionResult(action_id=action_id, status=status)

    def get_status(self) -> NodeStatus:
        """Get the status of the node."""
        response = requests.get(f"{self.url}/status", timeout=10)
        response.raise_for_status()
        return NodeStatus.model_validate(response.json())

    def get_state(self) -> dict[str, Any]:
        """Get the state of the node."""
        response = requests.get(f"{self.url}/state", timeout=10)
        response.raise_for_status()
        return response.json()

    def get_info(self) -> NodeInfo:
        """Get information about the node."""
        response = requests.get(f"{self.url}/info", timeout=10)
        response.raise_for_status()
        return NodeInfo.model_validate(response.json())

    def set_config(self, new_config: dict[str, Any]) -> NodeSetConfigResponse:
        """Update configuration values of the node."""
        response = requests.post(
            f"{self.url}/config",
            json=new_config,
            timeout=60,
        )
        response.raise_for_status()
        return NodeSetConfigResponse.model_validate(response.json())

    def send_admin_command(self, admin_command: AdminCommands) -> AdminCommandResponse:
        """Perform an administrative command on the node."""
        response = requests.post(f"{self.url}/admin/{admin_command}", timeout=10)
        response.raise_for_status()
        return AdminCommandResponse.model_validate(response.json())

    def get_resources(self) -> dict[str, ResourceDataModels]:
        """Get the resources of the node."""
        raise NotImplementedError(
            "get_resources is not implemented by this client",
        )
        # TODO: Implement get_resources endpoint

    def get_log(self) -> dict[str, Event]:
        """Get the log from the node"""
        response = requests.get(f"{self.url}/log", timeout=10)
        response.raise_for_status()
        return response.json()
